/*
 * interrupt handling is here:
 *  - ioapic control
 *  - lapic control
 *  - event channel management
 *  - irq handler
 */

#include "emu.h"

#include "cpufeature.h"
#include "msr-index.h"
#include "apicdef.h"
#include "ioapicdef.h"
#include "bitops.h"

/* --------------------------------------------------------------------- */

static void      *ioapic_mmio;
static uint32_t  ioapic_pins;

static uint32_t ioapic_read(int reg)
{
    uint32_t *sel = (ioapic_mmio + IOAPIC_REG_SELECT);
    uint32_t *win = (ioapic_mmio + IOAPIC_REG_WINDOW);
    *sel = reg;
    return *win;
}

static void ioapic_write(int reg, uint32_t val)
{
    uint32_t *sel = (ioapic_mmio + IOAPIC_REG_SELECT);
    uint32_t *win = (ioapic_mmio + IOAPIC_REG_WINDOW);
    *sel = reg;
    *win = val;
}

static void ioapic_write_entry(int pin, struct IO_APIC_route_entry e)
{
    union route_entry_union {
        struct { uint32_t w1, w2; };
        struct IO_APIC_route_entry entry;
    } eu;

    eu.entry = e;
    ioapic_write(0x11 + 2*pin, eu.w2);
    ioapic_write(0x10 + 2*pin, eu.w1);
}

static void ioapic_route_irq(int pin, int vector)
{
    struct IO_APIC_route_entry entry;

    memset(&entry,0,sizeof(entry));
    entry.vector = vector;

    ioapic_write_entry(pin, entry);
}

static void ioapic_init(void)
{
    ureg_t base = IOAPIC_DEFAULT_BASE_ADDRESS;
    uint32_t ver, id;

    ioapic_mmio = fixmap_page(base);
    id = ioapic_read(IOAPIC_REG_APIC_ID);
    ver = ioapic_read(IOAPIC_REG_VERSION);
    ioapic_pins = ((ver >> 16) & 0xff) + 1;
    printk(1, "%s: base %" PRIxREG ", mapped to %p, id %d, version %d, pins %d\n",
	   __FUNCTION__, base, ioapic_mmio,
	   (id >> 24) & 0x0f, ver & 0xff, ioapic_pins);
    if (0 == ver)
	panic("oops: ioapic version register is zero");

    /* PICs: mask all irqs */
    outb(0xff, 0x21);
    outb(0xff, 0xa1);
}

/* --------------------------------------------------------------------- */

static uint32_t lapic_read(struct xen_cpu *cpu, int reg)
{
    uint32_t *ptr = (cpu->lapic + reg);
    return *ptr;
}

static void lapic_write(struct xen_cpu *cpu, int reg, uint32_t val)
{
    uint32_t *ptr = (cpu->lapic + reg);
    *ptr = val;
}

static void lapic_eoi(struct xen_cpu *cpu)
{
    lapic_write(cpu, APIC_EOI, 0);
}

void lapic_timer(struct xen_cpu *cpu)
{
    uint64_t systime;
    uint32_t lvt;
    uint32_t count;
    uint32_t div;
    int64_t nsecs;

    systime = cpu->vcpu_info->time.system_time;
    if (cpu->oneshot) {
	nsecs = cpu->oneshot - systime;
	if (nsecs < 1)
	    nsecs = 1;
    } else {
	nsecs = cpu->periodic;
    }
    printk(2, "%s/%d: periodic %" PRId64 ", oneshot %" PRId64
	   ", systime %" PRId64 ", nsecs %" PRId64 "\n", __FUNCTION__,
	   cpu->id, cpu->periodic, cpu->oneshot, systime, nsecs);

    lvt = cpu->virq_to_vector[VIRQ_TIMER];
    if (!cpu->oneshot)
	lvt |= (1 << 17);

    div = APIC_TDR_DIV_1;
    count = nsecs;  /* kvm virtual apic has 1 ns ticks */
    if (count != nsecs) {
	/* count overflow, get some more bits */
	div = APIC_TDR_DIV_128;
	count = nsecs / 128;
	if (count != nsecs / 128) {
	    /* Hmm, still overflows ... */
	    printk(0, "%s: nsecs 0x%" PRIx64 ", nsecs/128 0x%" PRIx64 ", count 0x%x\n",
		   __FUNCTION__, nsecs, nsecs / 128, count);
	    panic("lapic timer count overflow");
	}
    }

    lapic_write(cpu, APIC_LVTT,  lvt);
    lapic_write(cpu, APIC_TDCR,  div);
    lapic_write(cpu, APIC_TMICT, count);
}

static int lapic_init(struct xen_cpu *cpu)
{
    struct kvm_cpuid_entry entry;
    uint32_t ax, dx, ver, id, spiv;
    ureg_t base;

    entry.function = 0x00000001;
    real_cpuid(&entry);
    if (!(entry.edx & (1 << (X86_FEATURE_APIC    %32)))) {
	printk(1, "%s: no lapic present\n", __FUNCTION__);
	return 0;
    }

    rdmsr(MSR_IA32_APICBASE, &ax, &dx);
    base = (uint64_t)dx << 32 | (ax & PAGE_MASK);
    cpu->lapic = fixmap_page(base);
    id   = lapic_read(cpu, APIC_ID);
    ver  = lapic_read(cpu, APIC_LVR);
    spiv = lapic_read(cpu, APIC_SPIV);
    printk(1, "%s: base %" PRIxREG ", mapped to %p, id %d, version %d, maxlvt %d%s%s%s\n",
	   __FUNCTION__, base, cpu->lapic,
	   GET_APIC_ID(id),
	   GET_APIC_VERSION(ver),
	   GET_APIC_MAXLVT(ver),
	   ax & 0x00000100 ? ", bootcpu"    : "",
	   ax & 0x00000800 ? ", hw-enabled" : "",
	   spiv & APIC_SPIV_APIC_ENABLED ? ", sw-enabled" : "");
    if (0 == ver)
	panic("oops: lapic version register is zero");

    lapic_write(cpu, APIC_SPIV, spiv | APIC_SPIV_APIC_ENABLED);

    if (ax & 0x00000100) /* boot cpu */
	return 1;
    return 2;
}

/* --------------------------------------------------------------------- */

static int vector_next = 0x30;

static struct vector {
    int  evtchn;
    int  virq;
    char *desc;
} vectors[256] = {
    [ 0 ... 255 ] = { .virq = -1 }
};

int evtchn_route_interdomain(int port, char *desc)
{
    int vec = vector_next++;

    ioapic_route_irq(port, vec);
    vectors[vec].evtchn = port;
    vectors[vec].desc   = desc ? desc : "no desc";
    printk(1, "%s: irq vector %d = event channel %d (%s)\n", __FUNCTION__,
	   vec, vectors[vec].evtchn, vectors[vec].desc);
    return 0;
}

int evtchn_route_virq(int virq, int port, char *desc)
{
    int vec = vector_next++;

    vectors[vec].evtchn = port;
    vectors[vec].virq   = virq;
    vectors[vec].desc   = desc ? desc : "no desc";
    printk(1, "%s: irq vector %d = event channel %d (virq %d, %s)\n", __FUNCTION__,
	   vec, vectors[vec].evtchn, vectors[vec].virq, vectors[vec].desc);
    return vec;
}

int evtchn_alloc(void)
{
    ureg_t rax;

    /* magic i/o port, on return rax is filled with
     * a fresh allocated event channel port */
    asm volatile("outb %%al,$0xee"
		 : "=a" (rax)
		 : /* no input */
		 : "memory");
    return rax;
}

static int evtchn_route_init(struct xen_cpu *cpu)
{
    struct start_info *si = (void*)(cpu->init_ctxt->user_regs.esi);

    evtchn_route_interdomain(si->console.domU.evtchn, "console");
    evtchn_route_interdomain(si->store_evtchn, "xenstore");
    return 0;
}

/* --------------------------------------------------------------------- */

static void evtchn_raise_event(struct xen_cpu *cpu, int port)
{
    int word = port / (sizeof(intptr_t)*8);

    if (test_and_set_bit(port, shared_info.evtchn_pending))
	return;
    if (test_bit(port, shared_info.evtchn_mask))
	return;
    if (test_and_set_bit(word, &cpu->vcpu_info->evtchn_pending_sel))
	return;
    cpu->vcpu_info->evtchn_upcall_pending = 1;
}

int evtchn_pending(struct xen_cpu *cpu)
{
    if (cpu->vcpu_info->evtchn_upcall_mask)
	return 0;
    if (!cpu->vcpu_info->evtchn_upcall_pending)
	return 0;
    return 1;
}

static void evtchn_forward(struct xen_cpu *cpu, struct regs *regs)
{
    vminfo.faults[XEN_FAULT_EVENT_CALLBACK]++;
    bounce_trap(cpu, regs, -1, CALLBACKTYPE_event);
#if longmode
    /* return via iretq please */
    regs->error = HCALL_IRET;
#endif
}

void evtchn_try_forward(struct xen_cpu *cpu, struct regs *regs)
{
    if (!evtchn_pending(cpu))
	return;
    evtchn_forward(cpu, regs);
}

/* --------------------------------------------------------------------- */

asmlinkage void do_irq(struct regs *regs)
{
    struct xen_cpu *cpu =get_cpu();

    lapic_eoi(cpu);
    switch (vectors[regs->trapno].virq) {
    case -1:
	/* nothing */
	break;
    case VIRQ_TIMER:
	if (cpu->oneshot) {
	    cpu->oneshot = 0;
	    lapic_timer(cpu);
	}
	pv_update_systime(0);
	break;
    }
    if (vectors[regs->trapno].evtchn) {
	printk(3, "%s: irq vector %d, evtchn %d (%s)\n", __FUNCTION__,
	       (int)regs->trapno,
	       vectors[regs->trapno].evtchn,
	       vectors[regs->trapno].desc);
	vminfo.faults[XEN_FAULT_EVTCHN_START + vectors[regs->trapno].evtchn]++;
	evtchn_raise_event(cpu, vectors[regs->trapno].evtchn);
	evtchn_try_forward(cpu, regs);
	return;
    }
    
    printk(0, "%s: unhandled irq (vector %d)\n", __FUNCTION__, (int)regs->trapno);
    print_registers(0, regs);
    panic("unknown irq");
}

int irq_init(struct xen_cpu *cpu)
{
    int rc;

    rc = lapic_init(cpu);
    if (0 == rc)
	return 0;
    if (1 == rc) {
	/* boot cpu */
	ioapic_init();
	evtchn_route_init(cpu);
    }
    return rc;
}
